@@ -1,6 +1,6 @@
 /*
  * This file is part of libaacs
- * Copyright (C) 2010  npzacs
+ * Copyright (C) 2010-2013  npzacs
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -19,22 +19,7 @@
 
 #include "keydbcfg.h"
 
-#if defined(__APPLE__)
-# include "darwin.h"
-# define get_config_home darwin_get_config_home
-# define get_cache_home darwin_get_cache_home
-# define get_config_system darwin_get_config_system
-#elif defined(_WIN32)
-# include "win32.h"
-# define get_config_home win32_get_config_home
-# define get_cache_home win32_get_config_home
-# define get_config_system win32_get_config_system
-#else
-# include "xdg.h"
-# define get_config_home xdg_get_config_home
-# define get_cache_home xdg_get_cache_home
-# define get_config_system xdg_get_config_system
-#endif
+#include "dirs.h"
 
 #include "util/strutl.h"
 #include "util/logging.h"
@@ -60,6 +45,36 @@
 #define MIN_FILE_SIZE  20
 #define MAX_FILE_SIZE  65535
 
+static int _mkpath(const char *path)
+{
+    struct stat s;
+    int result = 1;
+    char *dir = str_printf("%s", path);
+    char *end = dir;
+
+    while (*end == '/')
+        end++;
+
+    while ((end = strchr(end, '/'))) {
+        *end = 0;
+
+        if (stat(dir, &s) != 0 || !S_ISDIR(s.st_mode)) {
+            DEBUG(DBG_FILE, "Creating directory %s\n", dir);
+
+            if (mkdir(dir, S_IRWXU|S_IRWXG|S_IRWXO) == -1) {
+                DEBUG(DBG_FILE | DBG_CRIT, "Error creating directory %s\n", dir);
+                result = 0;
+                break;
+            }
+        }
+
+        *end++ = '/';
+    }
+
+    X_FREE(dir);
+
+    return result;
+}
 
 static char *_load_file(FILE *fp)
 {
@@ -89,18 +104,35 @@ static char *_load_file(FILE *fp)
     return data;
 }
 
-static FILE *_open_cfg_file_user(const char *file_name, char **path)
+static char *_config_file_user(const char *file_name)
 {
-    const char *cfg_dir = get_config_home();
+    const char *cfg_dir = file_get_config_home();
 
     if (!cfg_dir) {
         return NULL;
     }
 
-    char *cfg_file = str_printf("%s/%s/%s", cfg_dir, CFG_DIR, file_name);
-    FILE *fp       = fopen(cfg_file, "r");
+    return str_printf("%s/%s/%s", cfg_dir, CFG_DIR, file_name);
+}
 
-    DEBUG(DBG_FILE, fp ? "Reading %s\n" : "%s not found\n", cfg_file);
+static FILE *_open_cfg_file_user(const char *file_name, char **path, const char *mode)
+{
+    char *cfg_file = _config_file_user(file_name);
+
+    if (!cfg_file) {
+        return NULL;
+    }
+
+    if (*mode == 'w') {
+        if (!_mkpath(cfg_file)) {
+            X_FREE(cfg_file);
+            return NULL;
+        }
+    }
+
+    FILE *fp = fopen(cfg_file, mode);
+
+    DEBUG(DBG_FILE, fp ? "Opened %s for %s\n" : "%s not found\n", cfg_file, mode);
 
     if (fp && path) {
         *path = cfg_file;
@@ -115,7 +147,7 @@ static FILE *_open_cfg_file_system(const char *file_name, char **path)
 {
     const char *dir = NULL;
 
-    while (NULL != (dir = get_config_system(dir))) {
+    while (NULL != (dir = file_get_config_system(dir))) {
 
         char *cfg_file = str_printf("%s/%s/%s", dir, CFG_DIR, file_name);
 
@@ -139,13 +171,11 @@ static FILE *_open_cfg_file_system(const char *file_name, char **path)
     return NULL;
 }
 
-static int _is_duplicate_pk(pk_list *list, const char *e)
+static int _is_duplicate_pk(pk_list *list, const uint8_t *e)
 {
     while (list) {
-        if (list->key) {
-            if (!memcmp(list->key, e, 2*16)) {
-                return 1;
-            }
+        if (!memcmp(list->key, e, 16)) {
+            return 1;
         }
         list = list->next;
     }
@@ -164,22 +194,24 @@ static int _parse_pk_file(config_file *cf, FILE *fp)
         while (*p) {
             char *str = str_get_hex_string(p, 2*16);
 
-            if (str && _is_duplicate_pk(cf->pkl, str)) {
-                DEBUG(DBG_FILE, "Skipping duplicate processing key %s\n", str);
-                X_FREE(str);
-
-            } else if (str) {
+            if (str) {
                 DEBUG(DBG_FILE, "Found processing key %s\n", str);
 
                 pk_list *e = calloc(1, sizeof(pk_list));
 
-                e->key  = str;
-                e->next = cf->pkl;
+                hexstring_to_hex_array(e->key, 16, str);
 
-                cf->pkl = e;
+                if (_is_duplicate_pk(cf->pkl, e->key)) {
+                    DEBUG(DBG_FILE, "Skipping duplicate processing key %s\n", str);
+                    X_FREE(e);
+                } else {
+                    e->next = cf->pkl;
+                    cf->pkl = e;
+                }
 
                 result++;
             }
+            X_FREE(str);
 
             p = str_next_line(p);
         }
@@ -193,13 +225,10 @@ static int _parse_pk_file(config_file *cf, FILE *fp)
 static int _is_duplicate_cert(cert_list *list, cert_list *e)
 {
     while (list) {
-        if (list->host_priv_key && list->host_cert) {
+        if (!memcmp(list->host_priv_key, e->host_priv_key, 20) &&
+            !memcmp(list->host_cert,     e->host_cert,     92)) {
 
-            if (!memcmp(list->host_priv_key, e->host_priv_key, 2*20) &&
-                !memcmp(list->host_cert,     e->host_cert,     2*92)) {
-
-                return 1;
-            }
+            return 1;
         }
         list = list->next;
     }
@@ -207,13 +236,6 @@ static int _is_duplicate_cert(cert_list *list, cert_list *e)
   return 0;
 }
 
-static void _free_cert_entry(cert_list *e)
-{
-    X_FREE(e->host_priv_key);
-    X_FREE(e->host_cert);
-    X_FREE(e);
-}
-
 static int _parse_cert_file(config_file *cf, FILE *fp)
 {
     char *data   = _load_file(fp);
@@ -221,40 +243,59 @@ static int _parse_cert_file(config_file *cf, FILE *fp)
 
     if (data) {
         const char *p = data;
-        cert_list  *e = calloc(1, sizeof(cert_list));
 
-        e->host_priv_key = str_get_hex_string(p, 2*20);
+        char *host_cert, *host_priv_key;
+        host_priv_key = str_get_hex_string(p, 2*20);
         p = str_next_line(p);
-        e->host_cert = str_get_hex_string(p, 2*92);
+        host_cert = str_get_hex_string(p, 2*92);
 
         X_FREE(data);
 
-        if (!e->host_priv_key || !e->host_cert) {
+        if (!host_priv_key || !host_cert) {
             DEBUG(DBG_FILE, "Invalid file\n");
-            _free_cert_entry(e);
-
-        } else if (_is_duplicate_cert(cf->host_cert_list, e)) {
-            DEBUG(DBG_FILE, "Skipping duplicate certificate entry %s %s\n", e->host_priv_key, e->host_cert);
-            _free_cert_entry(e);
 
         } else {
-            DEBUG(DBG_FILE, "Found certificate: %s %s\n", e->host_priv_key, e->host_cert);
-            e->next = cf->host_cert_list;
-            cf->host_cert_list = e;
-            result = 1;
+            DEBUG(DBG_FILE, "Found certificate: %s %s\n", host_priv_key, host_cert);
+
+            cert_list  *e = calloc(1, sizeof(cert_list));
+            hexstring_to_hex_array(e->host_priv_key, 20, host_priv_key);
+            hexstring_to_hex_array(e->host_cert, 92, host_cert);
+
+            if (_is_duplicate_cert(cf->host_cert_list, e)) {
+                DEBUG(DBG_FILE, "Skipping duplicate certificate entry %s %s\n", host_priv_key, host_cert);
+                X_FREE(e);
+            } else {
+                e->next = cf->host_cert_list;
+                cf->host_cert_list = e;
+                result = 1;
+            }
         }
+        X_FREE(host_priv_key);
+        X_FREE(host_cert);
     }
 
     return result;
 }
 
+static int _is_duplicate_dk(dk_list *list, dk_list *e)
+{
+    while (list) {
+      if (!memcmp(list, e, sizeof(*e))) {
+            return 1;
+        }
+        list = list->next;
+    }
+
+    return 0;
+}
+
 static int _load_pk_file(config_file *cf)
 {
     static const char pk_file_name[] = PK_FILE_NAME;
     FILE *fp;
     int result = 0;
 
-    fp = _open_cfg_file_user(pk_file_name, NULL);
+    fp = _open_cfg_file_user(pk_file_name, NULL, "r");
     if (fp) {
         result += _parse_pk_file(cf, fp);
         fclose(fp);
@@ -275,7 +316,7 @@ static int _load_cert_file(config_file *cf)
     FILE *fp;
     int result = 0;
 
-    fp = _open_cfg_file_user(cert_file_name, NULL);
+    fp = _open_cfg_file_user(cert_file_name, NULL, "r");
     if (fp) {
         result += _parse_cert_file(cf, fp);
         fclose(fp);
@@ -290,40 +331,9 @@ static int _load_cert_file(config_file *cf)
     return result;
 }
 
-static int _mkpath(const char *path)
-{
-    struct stat s;
-    int result = 1;
-    char *dir = str_printf("%s", path);
-    char *end = dir;
-
-    while (*end == '/')
-        end++;
-
-    while ((end = strchr(end, '/'))) {
-        *end = 0;
-
-        if (stat(dir, &s) != 0 || !S_ISDIR(s.st_mode)) {
-            DEBUG(DBG_FILE, "Creating directory %s\n", dir);
-
-            if (mkdir(dir, S_IRWXU|S_IRWXG|S_IRWXO) == -1) {
-                DEBUG(DBG_FILE, "Error creating directory %s\n", dir);
-                result = 0;
-                break;
-            }
-        }
-
-        *end++ = '/';
-    }
-
-    X_FREE(dir);
-
-    return result;
-}
-
 static char *_keycache_file(const char *type, const uint8_t *disc_id)
 {
-    const char *cache_dir = get_cache_home();
+    const char *cache_dir = file_get_cache_home();
     char disc_id_str[41];
 
     if (!cache_dir) {
@@ -408,7 +418,7 @@ int keycache_find(const char *type, const uint8_t *disc_id, uint8_t *key, unsign
 
 static char *_cache_file(const char *name)
 {
-    const char *cache_dir = get_cache_home();
+    const char *cache_dir = file_get_cache_home();
 
     if (!cache_dir) {
         return NULL;
@@ -489,7 +499,7 @@ int cache_get(const char *name, uint32_t *version, uint32_t *len, void *buf)
 int cache_remove(const char *name)
 {
     char *file = _cache_file(name);
-    int result = !remove(name);
+    int result = !remove(file);
     if (!result) {
         DEBUG(DBG_FILE, "Error removing %s\n", file);
     }
@@ -533,6 +543,61 @@ void *cache_get_or_update(const char *type, const void *data, uint32_t *len, uin
     return cache_data;
 }
 
+int config_save(const char *name, const void *data, uint32_t len)
+{
+    char *path = NULL;
+    FILE *fp = _open_cfg_file_user(name, &path, "w");
+    int result = 0;
+
+    if (fp) {
+        if (fwrite(&len, 1, 4, fp) == 4 &&
+            fwrite(data, 1, len, fp) == len) {
+          DEBUG(DBG_FILE, "Wrote %d bytes to %s\n", len + 4, path);
+          result = 1;
+
+        } else {
+            DEBUG(DBG_FILE | DBG_CRIT, "Error writing to %s\n", path);
+        }
+
+        fclose(fp);
+    }
+
+    X_FREE(path);
+
+    return result;
+}
+
+int config_get(const char *name, uint32_t *len, void *buf)
+{
+    char *path = NULL;
+    FILE *fp = _open_cfg_file_user(name, &path, "r");
+    int result = 0;
+    uint32_t size = *len;
+
+    *len = 0;
+
+    if (fp) {
+        DEBUG(DBG_FILE, "Reading %s\n", path);
+
+        if (fread(len, 1, 4, fp) == 4 && (size <= *len) &&
+            (!buf || fread(buf, 1, *len, fp) == *len)) {
+
+            DEBUG(DBG_FILE, "Read %d bytes from %s\n", 4 + (buf ? *len : 0), path);
+            result = 1;
+
+        } else {
+            DEBUG(DBG_FILE | DBG_CRIT, "Error reading from %s\n", path);
+        }
+
+        fclose(fp);
+    }
+
+    X_FREE(path);
+
+    return result;
+}
+
+
 static char *_find_config_file(void)
 {
     static const char cfg_file_name[] = CFG_FILE_NAME;
@@ -540,7 +605,7 @@ static char *_find_config_file(void)
     char       *cfg_file = NULL;
     FILE       *fp       = NULL;
 
-    fp = _open_cfg_file_user(cfg_file_name, &cfg_file);
+    fp = _open_cfg_file_user(cfg_file_name, &cfg_file, "r");
     if (!fp) {
         fp = _open_cfg_file_system(cfg_file_name, &cfg_file);
     }
@@ -553,6 +618,66 @@ static char *_find_config_file(void)
     return cfg_file;
 }
 
+#include "keydb.h"
+
+static int _parse_embedded(config_file *cf)
+{
+    int result = 0, jj;
+    unsigned ii;
+
+    /* reverse order to maintain key positions (items are added to list head) */
+    for (jj = sizeof(internal_dk_list) / sizeof(internal_dk_list[0]) - 1; jj >= 0; --jj) {
+        dk_list *e = calloc(1, sizeof(dk_list));
+
+        decrypt_key(e->key, internal_dk_list[jj], 16);
+        e->node = internal_device_number;
+        e->uv   = MKINT_BE32(internal_dk_list[jj] + 16);
+        e->u_mask_shift = internal_dk_list[jj][20];
+
+        if (!e->uv || _is_duplicate_dk(cf->dkl, e)) {
+            X_FREE(e);
+
+        } else {
+            e->next = cf->dkl;
+            cf->dkl = e;
+            result++;
+        }
+    }
+
+    for (ii = 0; ii < sizeof(internal_pk_list) / sizeof(internal_pk_list[0]); ii++) {
+        pk_list *e = calloc(1, sizeof(pk_list));
+
+        decrypt_key(e->key, internal_pk_list[ii], 16);
+
+        if (_is_duplicate_pk(cf->pkl, e->key)) {
+            X_FREE(e);
+
+        } else {
+            e->next = cf->pkl;
+            cf->pkl = e;
+            result++;
+        }
+    }
+
+    for (ii = 0; ii < sizeof(internal_hc_list) / sizeof(internal_hc_list[0]); ii++) {
+        cert_list  *e = calloc(1, sizeof(cert_list));
+
+        decrypt_key(e->host_priv_key, internal_hc_list[ii],      20);
+        decrypt_key(e->host_cert,     internal_hc_list[ii] + 20, 92);
+
+        if (_is_duplicate_cert(cf->host_cert_list, e)) {
+            X_FREE(e);
+
+        } else {
+            e->next = cf->host_cert_list;
+            cf->host_cert_list = e;
+            result++;
+        }
+    }
+
+    return result;
+}
+
 config_file *keydbcfg_config_load(const char *configfile_path)
 {
     int config_ok = 0;
@@ -578,6 +703,9 @@ config_file *keydbcfg_config_load(const char *configfile_path)
     config_ok = _load_pk_file(cf)   || config_ok;
     config_ok = _load_cert_file(cf) || config_ok;
 
+    /* embedded keys */
+    config_ok = _parse_embedded(cf) || config_ok;
+
     if (!config_ok) {
         DEBUG(DBG_AACS | DBG_CRIT, "No valid AACS configuration files found\n");
         keydbcfg_config_file_close(cf);
