@@ -1,7 +1,7 @@
 /*
  * This file is part of libaacs
  * Copyright (C) 2009-2010  Obliter0n
- * Copyright (C) 2009-2010  npzacs
+ * Copyright (C) 2009-2013  npzacs
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -54,6 +54,8 @@ struct aacs {
     uint8_t   vid[16];
     /* PMSN is cached for BD-J */
     uint8_t   pmsn[16];
+    /* Media key is cached for BD+ */
+    uint8_t   mk[16];
 
     /* unit key for each CPS unit */
     uint32_t  num_uks;
@@ -65,10 +67,21 @@ struct aacs {
     /* title -> CPS unit mappings */
     uint32_t  num_titles;
     uint16_t *cps_units;  /* [0] = first play ; [1] = top menu ; [2] = title 1 ... */
+
+    /* bus encryption */
+    int       bee;        /* bus encryption enabled flag in content certificate */
+    int       bec;        /* bus encryption capable flag in drive certificate */
+    uint8_t   read_data_key[16];
+
+    /* AACS Online (BD-J) */
+    uint8_t   device_nonce[16];
+    uint8_t   device_binding_id[16];
 };
 
 static const uint8_t empty_key[] = "\x00\x00\x00\x00\x00\x00\x00\x00"
                                    "\x00\x00\x00\x00\x00\x00\x00\x00";
+static const uint8_t aacs_iv[]   = "\x0b\xa0\xf8\xdd\xfe\xa6\x1f\xb3"
+                                   "\xd8\xdf\x9f\x56\x6a\x05\x0f\x78";
 
 static int _validate_pk(const uint8_t *pk,
                         const uint8_t *cvalue, const uint8_t *uv, const uint8_t *vd,
@@ -118,38 +131,245 @@ static int _rl_verify_signature(const uint8_t *rl, size_t size)
     return crypto_aacs_verify_aacsla(rl + len, rl, len);
 }
 
-static void _update_drl(MKB *mkb)
+static void _save_rl(const char *name, uint32_t version, const uint8_t *version_rec,
+                     const uint8_t *rl_rec, size_t rl_len)
+{
+    int len     = MKINT_BE24(rl_rec - 3);
+    int entries = MKINT_BE32(rl_rec + 4); /* entries in first signature block */
+    if (len < 4 || !entries) {
+        DEBUG(DBG_AACS | DBG_CRIT, "ignoring empty %s\n", name);
+        return;
+    }
+
+    if (rl_rec && version_rec) {
+        rl_rec -= 4;
+        rl_len += 4;
+
+        uint8_t *data = malloc(12 + rl_len);
+        memcpy(data,      version_rec, 12);
+        memcpy(data + 12, rl_rec,      rl_len);
+        if (!_rl_verify_signature(data, rl_len + 12)) {
+          DEBUG(DBG_AACS | DBG_CRIT, "invalid %s signature, not using it\n", name);
+        } else {
+          cache_save(name, version, data, rl_len + 12);
+        }
+        X_FREE(data);
+    }
+}
+
+static void _update_rl(MKB *mkb)
 {
     uint32_t version = mkb_version(mkb);
     uint32_t cache_version;
+    size_t   rl_len;
 
     if (!cache_get("drl", &cache_version, NULL, NULL) || cache_version < version) {
-        size_t drl_len;
-        const uint8_t *drl_rec = mkb_drive_revokation_entries(mkb, &drl_len);
-        const uint8_t *v_rec   = mkb_type_and_version_record(mkb);
-
-        if (drl_rec && v_rec) {
-            drl_rec -= 4;
-            drl_len += 4;
-
-            uint8_t *data = malloc(12 + drl_len);
-            memcpy(data,      v_rec,   12);
-            memcpy(data + 12, drl_rec, drl_len);
-            if (!_rl_verify_signature(data, drl_len + 12)) {
-                DEBUG(DBG_AACS | DBG_CRIT, "invalid drive revocation list signature, not using it\n");
-            } else {
-                cache_save("drl", version, data, drl_len + 12);
-            }
-            X_FREE(data);
+        const uint8_t *version_rec = mkb_type_and_version_record(mkb);
+        const uint8_t *drl_rec     = mkb_drive_revokation_entries(mkb, &rl_len);
+        if (drl_rec && rl_len > 8) {
+            _save_rl("drl", version, version_rec, drl_rec, rl_len);
+        }
+    }
+    if (!cache_get("hrl", &cache_version, NULL, NULL) || cache_version < version) {
+        const uint8_t *version_rec = mkb_type_and_version_record(mkb);
+        const uint8_t *hrl_rec     = mkb_host_revokation_entries(mkb, &rl_len);
+        if (hrl_rec && rl_len > 8) {
+            _save_rl("hrl", version, version_rec, hrl_rec, rl_len);
         }
     }
 }
 
-static int _calc_mk(AACS *aacs, uint8_t *mk, pk_list *pkl)
+static uint32_t _calc_v_mask(uint32_t uv)
+{
+    uint32_t v_mask = 0xffffffff;
+
+    while (!(uv & ~v_mask)) {
+        v_mask <<= 1;
+    }
+
+    return v_mask;
+}
+
+static void _calc_pk(const uint8_t *dk, uint8_t *pk, uint32_t uv, uint32_t v_mask, uint32_t dev_key_v_mask)
+{
+    unsigned char left_child[16], right_child[16];
+
+    crypto_aesg3(dk, left_child, right_child, pk);
+
+    while (dev_key_v_mask != v_mask) {
+
+        int i;
+        for (i = 31; i >= 0; i--) {
+            if (!(dev_key_v_mask & (1ul << i))) {
+                break;
+            }
+        }
+
+        uint8_t curr_key[16];
+        if (!(uv & (1ul << i))) {
+            memcpy(curr_key, left_child, 16);
+        } else {
+            memcpy(curr_key, right_child, 16);
+        }
+
+        crypto_aesg3(curr_key, left_child, right_child, pk);
+
+        dev_key_v_mask = ((int) dev_key_v_mask) >> 1;
+    }
+
+    char str[40];
+    DEBUG(DBG_AACS, "Processing key: %s\n",  print_hex(str, pk, 16));
+}
+
+static dk_list *_find_dk(dk_list *dkl, uint32_t *p_dev_key_v_mask, uint32_t uv, uint32_t u_mask)
+{
+    uint32_t device_number = dkl->node;
+    uint32_t dev_key_uv, dev_key_u_mask, dev_key_v_mask;
+    unsigned key_idx = 0;
+
+    for (; dkl; dkl = dkl->next) {
+        if (device_number != dkl->node) {
+            /* wrong device */
+            continue;
+        }
+        if (!dkl->uv) {
+            continue;
+        }
+        key_idx++;
+        dev_key_uv     = dkl->uv;
+        dev_key_u_mask = 0xffffffff << dkl->u_mask_shift;
+        dev_key_v_mask = _calc_v_mask(dev_key_uv);
+
+        if ((u_mask == dev_key_u_mask) &&
+            ((uv & dev_key_v_mask) == (dev_key_uv & dev_key_v_mask))) {
+            break;
+        }
+    }
+
+    if (!dkl) {
+        DEBUG(DBG_AACS | DBG_CRIT, "could not find applying device key (device 0x%x)\n", device_number);
+    } else {
+        char str[128];
+        DEBUG(DBG_AACS, "Applying device key is #%d %s\n", key_idx, print_hex(str, dkl->key, 16));
+        DEBUG(DBG_AACS, "  UV: 0x%08x  U mask: 0x%08x  V mask: 0x%08x\n", dev_key_uv, dev_key_u_mask, dev_key_v_mask);
+        *p_dev_key_v_mask = dev_key_v_mask;
+    }
+
+    return dkl;
+}
+
+static int _calc_pk_mk(MKB *mkb, dk_list *dkl, uint8_t *mk)
+{
+    /* calculate processing key and media key using device keys */
+
+    const uint8_t *uvs, *cvalues;
+    unsigned num_uvs;
+    size_t len;
+    char str[128];
+
+    /* get mkb data */
+
+    uvs     = mkb_subdiff_records(mkb, &len);
+    cvalues = mkb_cvalues(mkb, &len);
+    num_uvs = len / 5;
+
+    if (num_uvs < 1) {
+        return AACS_ERROR_CORRUPTED_DISC;
+    }
+
+    /* loop over all known devices */
+
+    dk_list *dk_num;
+    uint32_t device_number = (uint32_t)-1;
+
+    for (dk_num = dkl; dk_num; dk_num = dk_num->next) {
+
+        /* find next device */
+
+        if (device_number == dk_num->node) {
+            continue;
+        }
+        device_number = dk_num->node;
+
+        /* find applying subset difference */
+
+        unsigned uvs_idx;
+        uint32_t u_mask, v_mask, uv;
+
+        for (uvs_idx = 0; uvs_idx < num_uvs; uvs_idx++) {
+            const uint8_t *p_uv = uvs + 1 + 5 * uvs_idx;
+            uint8_t u_mask_shift = p_uv[-1];
+
+            uv = MKINT_BE32(p_uv);
+            if (!uv) {
+                continue;
+            }
+
+            if (u_mask_shift & 0xc0) {
+                DEBUG(DBG_AACS | DBG_CRIT, "device 0x%x is revoked\n", device_number);
+                uvs_idx = num_uvs;
+
+            } else {
+
+                u_mask = 0xffffffff << u_mask_shift;
+                v_mask = _calc_v_mask(uv);
+
+                if (((device_number & u_mask) == (uv & u_mask)) && ((device_number & v_mask) != (uv & v_mask))) {
+                    break;
+                }
+            }
+        }
+
+        if (uvs_idx >= num_uvs) {
+            DEBUG(DBG_AACS | DBG_CRIT, "could not find applying subset-difference for device 0x%x\n", device_number);
+            /* try next device */
+            continue;
+        }
+
+        DEBUG(DBG_AACS, "Applying subset-difference for device 0x%x is #%d:\n", device_number, uvs_idx);
+        DEBUG(DBG_AACS,"  UV: 0x%08x  U mask: 0x%08x  V mask: 0x%08x\n", uv, u_mask, v_mask);
+
+        /* find applying device key */
+
+        uint32_t dev_key_v_mask = 0;
+        dk_list *dk;
+
+        dk = _find_dk(dk_num, &dev_key_v_mask, uv, u_mask);
+        if (!dk) {
+            /* try next device */
+            continue;
+        }
+
+        /* calculate processing key */
+
+        uint8_t pk[16];
+        _calc_pk(dk->key, pk, uv, v_mask, dev_key_v_mask);
+
+        /* calculate and verify media key */
+
+        if ( _validate_pk(pk,
+                          cvalues + uvs_idx * 16,
+                          uvs + 1 + uvs_idx * 5,
+                          mkb_mk_dv(mkb),
+                          mk)
+             == AACS_SUCCESS) {
+
+            DEBUG(DBG_AACS, "Media key: %s\n", print_hex(str, mk, 16));
+            return AACS_SUCCESS;
+        }
+
+        DEBUG(DBG_AACS | DBG_CRIT, "Processing key %s is invalid!\n", print_hex(str, pk, 16));
+
+        /* try next device */
+    }
+
+    return AACS_ERROR_NO_DK;
+}
+
+static int _calc_mk(AACS *aacs, uint8_t *mk, pk_list *pkl, dk_list *dkl)
 {
     int a, num_uvs = 0;
     size_t len;
-    uint8_t *buf = NULL;
     MKB *mkb = NULL;
     const uint8_t *rec, *uvs;
 
@@ -161,9 +381,18 @@ static int _calc_mk(AACS *aacs, uint8_t *mk, pk_list *pkl)
     DEBUG(DBG_AACS, "Calculate media key...\n");
 
     if ((mkb = mkb_open(aacs->path))) {
-        DEBUG(DBG_AACS, "Get UVS...\n");
-        _update_drl(mkb);
+
         aacs->mkb_version = mkb_version(mkb);
+        _update_rl(mkb);
+
+        /* try device keys first */
+        if (dkl && _calc_pk_mk(mkb, dkl, mk) == AACS_SUCCESS) {
+            memcpy(aacs->mk, mk, sizeof(aacs->mk));
+            mkb_close(mkb);
+            return AACS_SUCCESS;
+        }
+
+        DEBUG(DBG_AACS, "Get UVS...\n");
         uvs = mkb_subdiff_records(mkb, &len);
         rec = uvs;
         while (rec < uvs + len) {
@@ -176,26 +405,23 @@ static int _calc_mk(AACS *aacs, uint8_t *mk, pk_list *pkl)
         DEBUG(DBG_AACS, "Get cvalues...\n");
         rec = mkb_cvalues(mkb, &len);
 
-        for (; pkl && pkl->key; pkl = pkl->next) {
-                uint8_t pk[16];
-                hexstring_to_hex_array(pk, sizeof(pk), pkl->key);
+        for (; pkl; pkl = pkl->next) {
                 DEBUG(DBG_AACS, "Trying processing key...\n");
 
                 for (a = 0; a < num_uvs; a++) {
-                    if (AACS_SUCCESS == _validate_pk(pk, rec + a * 16, uvs + 1 + a * 5,
+                    if (AACS_SUCCESS == _validate_pk(pkl->key, rec + a * 16, uvs + 1 + a * 5,
                       mkb_mk_dv(mkb), mk)) {
                         mkb_close(mkb);
-                        X_FREE(buf);
 
                         char str[40];
                         DEBUG(DBG_AACS, "Media key: %s\n", print_hex(str, mk, 16));
+                        memcpy(aacs->mk, mk, sizeof(aacs->mk));
                         return AACS_SUCCESS;
                     }
                 }
             }
 
         mkb_close(mkb);
-        X_FREE(buf);
 
         DEBUG(DBG_AACS | DBG_CRIT, "Error calculating media key. Missing right processing key ?\n");
         return AACS_ERROR_NO_PK;
@@ -231,7 +457,7 @@ static MKB *_get_hrl_mkb(MMC *mmc)
         data = cache_get_or_update("hrl", mkb_data(mkb), &size, mkb_version(mkb));
 
         if (!_rl_verify_signature(data, size)) {
-            DEBUG(DBG_AACS | DBG_CRIT, "invalid cached revocation list signature, replacing it\n");
+            DEBUG(DBG_AACS | DBG_CRIT, "invalid cached host revocation list signature, replacing it\n");
             cache_save("hrl", mkb_version(mkb), mkb_data(mkb), mkb_data_size(mkb));
             X_FREE(data);
         } else {
@@ -245,10 +471,10 @@ static MKB *_get_hrl_mkb(MMC *mmc)
         uint32_t size;
         data = cache_get_or_update("hrl", NULL, &size, 0);
         if (data && size > 0) {
-            if (!_rl_verify_signature(data, size)) {
+            if (_rl_verify_signature(data, size)) {
                 mkb = mkb_init(data, size);
             } else {
-                DEBUG(DBG_AACS | DBG_CRIT, "invalid cached revocation list signature, deleting cache\n");
+                DEBUG(DBG_AACS | DBG_CRIT, "invalid cached host revocation list signature, deleting cache\n");
                 cache_remove("hrl");
             }
         }
@@ -262,13 +488,8 @@ static MKB *_get_hrl_mkb(MMC *mmc)
     return mkb;
 }
 
-static int _read_vid(AACS *aacs, cert_list *hcl)
+static int _mmc_read_auth(AACS *aacs, cert_list *hcl, int type, uint8_t *p1, uint8_t *p2)
 {
-    /* Use VID given in config file if available */
-    if (memcmp(aacs->vid, empty_key, 16)) {
-        return AACS_SUCCESS;
-    }
-
     MMC* mmc = NULL;
     if (!(mmc = mmc_open(aacs->path))) {
         return AACS_ERROR_MMC_OPEN;
@@ -277,31 +498,35 @@ static int _read_vid(AACS *aacs, cert_list *hcl)
     int error_code = AACS_ERROR_NO_CERT;
 
     MKB *hrl_mkb = _get_hrl_mkb(mmc);
+    const uint8_t *drive_cert = mmc_get_drive_cert(mmc);
 
-    for (;hcl && hcl->host_priv_key && hcl->host_cert; hcl = hcl->next) {
+    for (; hcl ; hcl = hcl->next) {
 
         char tmp_str[2*92+1];
-        uint8_t priv_key[20], cert[92];
-        hexstring_to_hex_array(priv_key, sizeof(priv_key), hcl->host_priv_key);
-        hexstring_to_hex_array(cert,     sizeof(cert),     hcl->host_cert);
 
-        if (!crypto_aacs_verify_host_cert(cert)) {
+        if (!crypto_aacs_verify_host_cert(hcl->host_cert)) {
             DEBUG(DBG_AACS, "Not using invalid host certificate %s.\n",
-                  print_hex(tmp_str, cert, 92));
+                  print_hex(tmp_str, hcl->host_cert, 92));
             continue;
         }
 
-        if (mkb_host_cert_is_revoked(hrl_mkb, cert + 4)) {
+        if (mkb_host_cert_is_revoked(hrl_mkb, hcl->host_cert + 4) > 0) {
             DEBUG(DBG_AACS | DBG_CRIT, "Host certificate %s has been revoked.\n",
-                  print_hex(tmp_str, cert + 4, 6));
+                  print_hex(tmp_str, hcl->host_cert + 4, 6));
             error_code = AACS_ERROR_CERT_REVOKED;
             //continue;
         }
 
-        DEBUG(DBG_AACS, "Trying host certificate (id 0x%s)...\n",
-              print_hex(tmp_str, cert + 4, 6));
+        if (drive_cert && (drive_cert[1] & 0x01) && !(hcl->host_cert[1] & 0x01)) {
+            DEBUG(DBG_AACS, "Certificate (id 0x%s) does not support bus encryption\n",
+                  print_hex(tmp_str, hcl->host_cert + 4, 6));
+            //continue;
+        }
 
-        int mmc_result = mmc_read_vid(mmc, priv_key, cert, aacs->vid, aacs->pmsn);
+        DEBUG(DBG_AACS, "Trying host certificate (id 0x%s)...\n",
+              print_hex(tmp_str, hcl->host_cert + 4, 6));
+
+        int mmc_result = mmc_read_auth(mmc, hcl->host_priv_key, hcl->host_cert, type, p1, p2);
         switch (mmc_result) {
             case MMC_SUCCESS:
                 mkb_close(hrl_mkb);
@@ -320,12 +545,45 @@ static int _read_vid(AACS *aacs, cert_list *hcl)
     mkb_close(hrl_mkb);
     mmc_close(mmc);
 
-    DEBUG(DBG_AACS, "Error reading VID!\n");
     return error_code;
 }
 
-static int _calc_vuk(AACS *aacs, uint8_t *mk, uint8_t *vuk,
-                     pk_list *pkl, cert_list *host_cert_list)
+static int _read_vid(AACS *aacs, cert_list *hcl)
+{
+    /* Use VID given in config file if available */
+    if (memcmp(aacs->vid, empty_key, 16)) {
+        return AACS_SUCCESS;
+    }
+
+    int error_code = _mmc_read_auth(aacs, hcl, MMC_READ_VID, aacs->vid, NULL);
+    if (error_code != AACS_SUCCESS) {
+        DEBUG(DBG_AACS, "Error reading VID!\n");
+    } else {
+        /* cache vid */
+        keycache_save("vid", aacs->disc_id, aacs->vid, 16);
+    }
+    return error_code;
+}
+
+static int _read_read_data_key(AACS *aacs, cert_list *hcl)
+{
+    int error_code = _mmc_read_auth(aacs, hcl, MMC_READ_DATA_KEYS, aacs->read_data_key, NULL);
+    if (error_code != AACS_SUCCESS) {
+        DEBUG(DBG_AACS, "Error reading data keys!\n");
+    }
+    return error_code;
+}
+
+static int _read_pmsn(AACS *aacs, cert_list *hcl)
+{
+    int error_code = _mmc_read_auth(aacs, hcl, MMC_READ_PMSN, aacs->pmsn, NULL);
+    if (error_code != AACS_SUCCESS) {
+        DEBUG(DBG_AACS, "Error reading PMSN!\n");
+    }
+    return error_code;
+}
+
+static int _calc_vuk(AACS *aacs, uint8_t *mk, uint8_t *vuk, config_file *cf)
 {
     int error_code;
 
@@ -341,26 +599,27 @@ static int _calc_vuk(AACS *aacs, uint8_t *mk, uint8_t *vuk,
         return AACS_SUCCESS;
     }
 
+    if (!cf) {
+        return AACS_ERROR_NO_CONFIG;
+    }
+
     /* make sure we have media key */
-    error_code = _calc_mk(aacs, mk, pkl);
+    error_code = _calc_mk(aacs, mk, cf->pkl, cf->dkl);
     if (error_code != AACS_SUCCESS) {
         return error_code;
     }
 
     /* acquire VID */
-    error_code = _read_vid(aacs, host_cert_list);
+    error_code = _read_vid(aacs, cf->host_cert_list);
     if (error_code != AACS_SUCCESS) {
         return error_code;
     }
 
+    /* calculate VUK */
+
+    crypto_aes128d(mk, aacs->vid, vuk);
+
     int a;
-    gcry_cipher_hd_t gcry_h;
-
-    gcry_cipher_open(&gcry_h, GCRY_CIPHER_AES, GCRY_CIPHER_MODE_ECB, 0);
-    gcry_cipher_setkey(gcry_h, mk, 16);
-    gcry_cipher_decrypt(gcry_h, vuk, 16, aacs->vid, 16);
-    gcry_cipher_close(gcry_h);
-
     for (a = 0; a < 16; a++) {
         vuk[a] ^= aacs->vid[a];
     }
@@ -439,23 +698,35 @@ static AACS_FILE_H *_open_unit_key_file(const char *path)
     return fp;
 }
 
+static AACS_FILE_H *_open_content_certificate_file(const char *path)
+{
+    AACS_FILE_H *fp;
+    char        *f_name;
+
+    f_name = str_printf("%s/AACS/Content000.cer", path);
+    fp = file_open(f_name, "rb");
+
+    if (!fp) {
+        DEBUG(DBG_AACS | DBG_CRIT, "Error opening content certificate file %s\n", f_name);
+    }
+
+    X_FREE(f_name);
+    return fp;
+}
+
 /* Function that collects keys from keydb config entry */
 static void _find_config_entry(AACS *aacs, title_entry_list *ce,
                                uint8_t *mk, uint8_t *vuk)
 {
-    uint8_t discid[20];
     char str[48];
 
     aacs->uks = NULL;
     aacs->num_uks = 0;
 
         while (ce && ce->entry.discid) {
-            memset(discid, 0, sizeof(discid));
-            hexstring_to_hex_array(discid, sizeof(discid),
-                                   ce->entry.discid);
-            if (!memcmp(aacs->disc_id, discid, 20)) {
+            if (!memcmp(aacs->disc_id, ce->entry.discid, 20)) {
                 DEBUG(DBG_AACS, "Found config entry for discid %s\n",
-                      ce->entry.discid);
+                      print_hex(str, ce->entry.discid, 20));
                 break;
             }
 
@@ -507,7 +778,7 @@ static void _find_config_entry(AACS *aacs, title_entry_list *ce,
         }
 }
 
-static int _calc_uks(AACS *aacs, const char *configfile_path)
+static int _calc_uks(AACS *aacs, config_file *cf)
 {
     AACS_FILE_H *fp = NULL;
     uint8_t  buf[16];
@@ -515,26 +786,20 @@ static int _calc_uks(AACS *aacs, const char *configfile_path)
     unsigned int i;
     int error_code;
 
-    config_file *cf = NULL;
     uint8_t mk[16] = {0}, vuk[16] = {0};
 
-    cf = keydbcfg_config_load(configfile_path);
-    if (!cf) {
-        return AACS_ERROR_NO_CONFIG;
-    }
+    if (cf) {
+        DEBUG(DBG_AACS, "Searching for keydb config entry...\n");
+        _find_config_entry(aacs, cf->list, mk, vuk);
 
-    DEBUG(DBG_AACS, "Searching for keydb config entry...\n");
-    _find_config_entry(aacs, cf->list, mk, vuk);
-
-    /* Skip if retrieved from config file */
-    if (aacs->uks) {
-        keydbcfg_config_file_close(cf);
-        return AACS_SUCCESS;
+        /* Skip if retrieved from config file */
+        if (aacs->uks) {
+            return AACS_SUCCESS;
+        }
     }
 
     /* Make sure we have VUK */
-    error_code = _calc_vuk(aacs, mk, vuk, cf->pkl, cf->host_cert_list);
-    keydbcfg_config_file_close(cf);
+    error_code = _calc_vuk(aacs, mk, vuk, cf);
     if (error_code != AACS_SUCCESS) {
         return error_code;
     }
@@ -577,12 +842,7 @@ static int _calc_uks(AACS *aacs, const char *configfile_path)
                 break;
             }
 
-            gcry_cipher_hd_t gcry_h;
-            gcry_cipher_open(&gcry_h, GCRY_CIPHER_AES,
-                             GCRY_CIPHER_MODE_ECB, 0);
-            gcry_cipher_setkey(gcry_h, vuk, 16);
-            gcry_cipher_decrypt(gcry_h, aacs->uks + 16*i, 16, buf, 16);
-            gcry_cipher_close(gcry_h);
+            crypto_aes128d(vuk, buf, aacs->uks + 16*i);
 
             char str[40];
             DEBUG(DBG_AACS, "Unit key %d: %s\n", i,
@@ -628,7 +888,7 @@ static int _calc_title_hash(const char *path, uint8_t *title_hash)
 
     } else {
         result = AACS_ERROR_CORRUPTED_DISC;
-        DEBUG(DBG_AACS | DBG_CRIT, "Failed to read %"PRIu64" bytes from unit key file %s/AACS/Unit_Key_RO.inf", f_size, path);
+        DEBUG(DBG_AACS | DBG_CRIT, "Failed to read %lu bytes from unit key file %s/AACS/Unit_Key_RO.inf", (unsigned long)f_size, path);
     }
 
     file_close(fp);
@@ -637,6 +897,50 @@ static int _calc_title_hash(const char *path, uint8_t *title_hash)
     return result;
 }
 
+static int _get_bus_encryption_enabled(const char *path)
+{
+    AACS_FILE_H *fp = NULL;
+    uint8_t buf[2];
+    int bee = 0;
+
+    fp = _open_content_certificate_file(path);
+    if (!fp) {
+        DEBUG(DBG_AACS | DBG_CRIT, "Unable to open content certificate\n");
+        return 0;
+    }
+
+    if (file_read(fp, buf, 2) == 2) {
+        bee = (buf[1] & 0x80) >> 7;
+        DEBUG(DBG_AACS, "Bus Encryption Enabled flag in content certificate: %d\n", bee);
+    } else {
+        DEBUG(DBG_AACS | DBG_CRIT, "Failed to read Bus Encryption Enabled flag from content certificate file\n");
+    }
+
+    file_close(fp);
+    return bee;
+}
+
+static int _get_bus_encryption_capable(const char *path)
+{
+    MMC* mmc = NULL;
+    uint8_t drive_cert[92];
+    int bec = 0;
+
+    if (!(mmc = mmc_open(path))) {
+        return 0;
+    }
+
+    if (mmc_read_drive_cert(mmc, drive_cert) == MMC_SUCCESS) {
+        bec = drive_cert[1] & 1;
+        DEBUG(DBG_AACS, "Bus Encryption Capable flag in drive certificate: %d\n", bec);
+    } else {
+        DEBUG(DBG_AACS | DBG_CRIT, "Unable to read drive certificate\n");
+    }
+
+    mmc_close(mmc);
+    return bec;
+}
+
 static int _verify_ts(uint8_t *buf, size_t size)
 {
     uint8_t *ptr;
@@ -673,8 +977,7 @@ static int _decrypt_unit(AACS *aacs, uint8_t *out_buf, const uint8_t *in_buf, ui
 {
     gcry_cipher_hd_t gcry_h;
     int a;
-    uint8_t key[16], iv[] = "\x0b\xa0\xf8\xdd\xfe\xa6\x1f\xb3"
-                            "\xd8\xdf\x9f\x56\x6a\x05\x0f\x78";
+    uint8_t key[16];
 
     gcry_cipher_open(&gcry_h, GCRY_CIPHER_AES, GCRY_CIPHER_MODE_ECB, 0);
     gcry_cipher_setkey(gcry_h, aacs->uks + curr_uk * 16, 16);
@@ -689,7 +992,7 @@ static int _decrypt_unit(AACS *aacs, uint8_t *out_buf, const uint8_t *in_buf, ui
 
     gcry_cipher_open(&gcry_h, GCRY_CIPHER_AES, GCRY_CIPHER_MODE_CBC, 0);
     gcry_cipher_setkey(gcry_h, key, 16);
-    gcry_cipher_setiv(gcry_h, iv, 16);
+    gcry_cipher_setiv(gcry_h, aacs_iv, 16);
     gcry_cipher_decrypt(gcry_h, out_buf + 16, ALIGNED_UNIT_LEN - 16, in_buf + 16, ALIGNED_UNIT_LEN - 16);
     gcry_cipher_close(gcry_h);
 
@@ -704,6 +1007,18 @@ static int _decrypt_unit(AACS *aacs, uint8_t *out_buf, const uint8_t *in_buf, ui
     return 0;
 }
 
+#define SECTOR_LEN 2048
+static void _decrypt_bus(AACS *aacs, uint8_t *buf)
+{
+    gcry_cipher_hd_t gcry_h;
+
+    gcry_cipher_open(&gcry_h, GCRY_CIPHER_AES, GCRY_CIPHER_MODE_CBC, 0);
+    gcry_cipher_setkey(gcry_h, aacs->read_data_key, 16);
+    gcry_cipher_setiv(gcry_h, aacs_iv, 16);
+    gcry_cipher_decrypt(gcry_h, buf + 16, SECTOR_LEN - 16, NULL, 0);
+    gcry_cipher_close(gcry_h);
+}
+
 void aacs_get_version(int *major, int *minor, int *micro)
 {
     *major = AACS_VERSION_MAJOR;
@@ -718,7 +1033,7 @@ AACS *aacs_open(const char *path, const char *configfile_path)
     AACS *aacs;
 
     aacs = aacs_open2(path, configfile_path, &error_code);
-    if (error_code == AACS_SUCCESS) {
+    if (!aacs || error_code == AACS_SUCCESS) {
         return aacs;
     }
 
@@ -728,7 +1043,7 @@ AACS *aacs_open(const char *path, const char *configfile_path)
 
 AACS *aacs_open2(const char *path, const char *configfile_path, int *error_code)
 {
-    DEBUG(DBG_AACS, "libaacs "AACS_VERSION_STRING" [%zd]\n", sizeof(AACS));
+    DEBUG(DBG_AACS, "libaacs "AACS_VERSION_STRING" [%u]\n", (unsigned)sizeof(AACS));
 
     DEBUG(DBG_AACS, "Initializing libgcrypt...\n");
     if (!crypto_init()) {
@@ -737,6 +1052,7 @@ AACS *aacs_open2(const char *path, const char *configfile_path, int *error_code)
     }
 
     AACS *aacs = calloc(1, sizeof(AACS));
+    config_file *cf;
 
     aacs->path = str_printf("%s", path);
 
@@ -746,15 +1062,34 @@ AACS *aacs_open2(const char *path, const char *configfile_path, int *error_code)
         return NULL;
     }
 
-    DEBUG(DBG_AACS, "Starting AACS waterfall...\n");
-    *error_code = _calc_uks(aacs, configfile_path);
+    cf = keydbcfg_config_load(configfile_path);
 
-    if (*error_code == AACS_SUCCESS) {
-        DEBUG(DBG_AACS, "AACS initialized!\n");
-    } else {
+    DEBUG(DBG_AACS, "Starting AACS waterfall...\n");
+    *error_code = _calc_uks(aacs, cf);
+    if (*error_code != AACS_SUCCESS) {
         DEBUG(DBG_AACS, "Failed to initialize AACS!\n");
     }
 
+    aacs->bee = _get_bus_encryption_enabled(path);
+    aacs->bec = _get_bus_encryption_capable(path);
+
+    if (*error_code == AACS_SUCCESS && aacs->bee && aacs->bec) {
+
+        if (!cf) {
+            *error_code = AACS_ERROR_NO_CONFIG;
+            return aacs;
+        }
+
+        *error_code = _read_read_data_key(aacs, cf->host_cert_list);
+        if (*error_code != AACS_SUCCESS) {
+            DEBUG(DBG_AACS | DBG_CRIT, "Unable to initialize bus encryption required by drive and disc\n");
+        }
+    }
+
+    keydbcfg_config_file_close(cf);
+
+    DEBUG(DBG_AACS, "AACS initialized!\n");
+
     return aacs;
 }
 
@@ -763,29 +1098,43 @@ void aacs_close(AACS *aacs)
     if (!aacs)
         return;
 
+    /* erase sensitive data */
+    if (aacs->uks) {
+        memset(aacs->uks, 0, 16 * aacs->num_uks);
+    }
+
     X_FREE(aacs->uks);
     X_FREE(aacs->cps_units);
     X_FREE(aacs->path);
 
     DEBUG(DBG_AACS, "AACS destroyed!\n");
 
+    /* erase sensitive data */
+    memset(aacs, 0, sizeof(*aacs));
+
     X_FREE(aacs);
 }
 
 int aacs_decrypt_unit(AACS *aacs, uint8_t *buf)
 {
     uint8_t out_buf[ALIGNED_UNIT_LEN];
+    int i;
 
     if (!(buf[0] & 0xc0)) {
         // TP_extra_header Copy_permission_indicator == 0, unit is not encrypted
         return 1;
     }
 
+    if (aacs->bee && aacs->bec) {
+        for (i = 0; i < ALIGNED_UNIT_LEN; i += SECTOR_LEN) {
+            _decrypt_bus(aacs, buf + i);
+        }
+    }
+
     if (_decrypt_unit(aacs, out_buf, buf, aacs->current_cps_unit)) {
         memcpy(buf, out_buf, ALIGNED_UNIT_LEN);
 
         // Clear copy_permission_indicator bits
-        int i;
         for (i = 0; i < 6144; i += 192) {
             buf[i] &= ~0xc0;
         }
@@ -815,32 +1164,94 @@ const uint8_t *aacs_get_disc_id(AACS *aacs)
     return aacs->disc_id;
 }
 
+const uint8_t *aacs_get_mk(AACS *aacs)
+{
+    if (!memcmp(aacs->mk, empty_key, 16)) {
+        config_file *cf = keydbcfg_config_load(NULL);
+        if (cf) {
+            _calc_mk(aacs, aacs->mk, cf->pkl, cf->dkl);
+
+            keydbcfg_config_file_close(cf);
+        }
+
+        if (!memcmp(aacs->mk, empty_key, 16)) {
+            DEBUG(DBG_AACS | DBG_CRIT, "aacs_get_mk() failed\n");
+            return NULL;
+        }
+    }
+
+    return aacs->mk;
+}
+
 const uint8_t *aacs_get_vid(AACS *aacs)
 {
     if (!memcmp(aacs->vid, empty_key, 16)) {
+        /* get cached vid */
+        if (keycache_find("vid", aacs->disc_id, aacs->vid, 16)) {
+            DEBUG(DBG_AACS, "Using cached VID\n");
+            return aacs->vid;
+        }
+
         config_file *cf = keydbcfg_config_load(NULL);
         if (cf) {
             _read_vid(aacs, cf->host_cert_list);
 
             keydbcfg_config_file_close(cf);
         }
+
+        if (!memcmp(aacs->vid, empty_key, 16)) {
+            DEBUG(DBG_AACS | DBG_CRIT, "aacs_get_vid() failed\n");
+            return NULL;
+        }
     }
+
     return aacs->vid;
 }
 
 const uint8_t *aacs_get_pmsn(AACS *aacs)
 {
-    if (!memcmp(aacs->vid, empty_key, 16)) {
+    if (!memcmp(aacs->pmsn, empty_key, 16)) {
         config_file *cf = keydbcfg_config_load(NULL);
         if (cf) {
-            _read_vid(aacs, cf->host_cert_list);
+            _read_pmsn(aacs, cf->host_cert_list);
 
             keydbcfg_config_file_close(cf);
         }
+
+        if (!memcmp(aacs->pmsn, empty_key, 16)) {
+            DEBUG(DBG_AACS, "aacs_get_pmsn() failed\n");
+            return NULL;
+        }
     }
+
     return aacs->pmsn;
 }
 
+const uint8_t *aacs_get_device_binding_id(AACS *aacs)
+{
+  /* Device binding ID is used to encrypt online content.
+   * It needs to be cached so that downloaded content can be played later.
+   */
+    static const char config_file_name[] = "device_binding_id";
+    uint32_t len = sizeof(aacs->device_binding_id);
+
+    DEBUG(DBG_AACS, "reading device binding id\n");
+    if (!config_get(config_file_name, &len, aacs->device_binding_id) || len != sizeof(aacs->device_binding_id)) {
+        DEBUG(DBG_AACS, "creating device binding id\n");
+        crypto_create_nonce(aacs->device_binding_id, sizeof(aacs->device_binding_id));
+        config_save(config_file_name, aacs->device_binding_id, sizeof(aacs->device_binding_id));
+    }
+
+   return aacs->device_binding_id;
+}
+
+const uint8_t *aacs_get_device_nonce(AACS *aacs)
+{
+    DEBUG(DBG_AACS, "creating device nonce\n");
+    crypto_create_nonce(aacs->device_nonce, sizeof(aacs->device_nonce));
+    return aacs->device_nonce;
+}
+
 static AACS_RL_ENTRY *_get_rl(const char *type, int *num_records, int *mkbv)
 {
     uint32_t len, version;
@@ -887,6 +1298,18 @@ AACS_RL_ENTRY *aacs_get_drl(int *num_records, int *mkbv)
     return _get_rl("drl", num_records, mkbv);
 }
 
+void aacs_free_rl(AACS_RL_ENTRY **rl)
+{
+    X_FREE(*rl);
+}
+
+uint32_t aacs_get_bus_encryption(AACS *aacs)
+{
+  return (aacs->bee * AACS_BUS_ENCRYPTION_ENABLED) |
+         (aacs->bec * AACS_BUS_ENCRYPTION_CAPABLE);
+}
+
+
 void aacs_select_title(AACS *aacs, uint32_t title)
 {
     if (!aacs) {
