@@ -76,7 +76,7 @@ MKB *mkb_open(const char *path)
 
         file_read(fp, mkb->buf, mkb->size);
 
-        DEBUG(DBG_MKB, "MKB size: %zd\n", mkb->size);
+        DEBUG(DBG_MKB, "MKB size: %u\n", (unsigned)mkb->size);
         DEBUG(DBG_MKB, "MKB version: %d\n", mkb_version(mkb));
 
         file_close(fp);
@@ -85,6 +85,8 @@ MKB *mkb_open(const char *path)
 
     DEBUG(DBG_MKB, "Error opening MKB!\n");
 
+    X_FREE(mkb);
+
     return NULL;
 }
 
@@ -221,7 +223,9 @@ static int _cert_is_revoked(const uint8_t *rl, size_t rl_size, const uint8_t *ce
             uint16_t len = MKINT_BE16(rl);
             uint64_t id  = MKINT_BE48(rl + 2);
             if (cert_id >= id && cert_id <= id + len) {
-                DEBUG(DBG_MKB, "Certificate %12llx has been revoked\n", (unsigned long long)cert_id);
+                char str[16];
+                print_hex(str, rl + 2, 6);
+                DEBUG(DBG_MKB, "Certificate %s has been revoked\n", str);
                 return 1;
             }
             rl += 8;
