@@ -28,8 +28,12 @@
 typedef struct mkb MKB;
 
 AACS_PRIVATE MKB *mkb_open(const char *path);    // init MKB
+AACS_PRIVATE MKB *mkb_init(uint8_t *data, int len); // init MKB from data
 AACS_PRIVATE void mkb_close(MKB *mkb);           // free MKB
 
+AACS_PRIVATE const uint8_t *mkb_data(MKB *mkb);
+AACS_PRIVATE size_t mkb_data_size(MKB *mkb);
+
 // returns type
 AACS_PRIVATE uint8_t mkb_type(MKB *mkb);
 // returns version
@@ -51,6 +55,9 @@ AACS_PRIVATE const uint8_t *mkb_mk_dv(MKB *mkb);
 // returns MKB signature
 AACS_PRIVATE const uint8_t *mkb_signature(MKB *mkb, size_t *len);
 
+AACS_PRIVATE int mkb_host_cert_is_revoked(MKB *mkb, const uint8_t *cert_id);
+AACS_PRIVATE int mkb_drive_cert_is_revoked(MKB *mkb, const uint8_t *cert_id);
+
 /* The Media Key block is of variable size but must be a multiple of 4
  * MKB Structure:
  *      Type and Version record (12 bytes)
