@@ -31,10 +31,16 @@ typedef struct mmc MMC;
 #define MMC_ERROR              -1 /* MMC failed */
 #define MMC_ERROR_CERT_REVOKED -2 /* revoked certificate */
 
+#define MMC_READ_VID            0
+#define MMC_READ_PMSN           1
+#define MMC_READ_DATA_KEYS      2
+
 AACS_PRIVATE MMC *mmc_open(const char *path);
 AACS_PRIVATE void mmc_close(MMC *mmc);
-AACS_PRIVATE int  mmc_read_vid(MMC *mmc, const uint8_t *host_priv_key, const uint8_t *host_cert,
-                               uint8_t *vid, uint8_t *pmsn);
+AACS_PRIVATE int  mmc_read_auth(MMC *mmc, const uint8_t *host_priv_key, const uint8_t *host_cert,
+                                int request, uint8_t *p1, uint8_t *p2);
+AACS_PRIVATE int  mmc_read_drive_cert(MMC *mmc, uint8_t *drive_cert);
+AACS_PRIVATE const uint8_t *mmc_get_drive_cert(MMC *mmc);
 
 /* read partial MKB */
 AACS_PRIVATE uint8_t *mmc_read_mkb(MMC *mmc, int address, int *size);
