@@ -24,7 +24,7 @@
 #include <stdarg.h>
 #include <string.h>
 
-static debug_mask_t debug_mask = DBG_CRIT;
+uint32_t debug_mask = (uint32_t)-1; /* set all bits to make sure aacs_debug() is called for initialization */
 
 char *print_hex(char *out, const uint8_t *buf, int count)
 {
@@ -45,6 +45,7 @@ void aacs_debug(const char *file, int line, uint32_t mask, const char *format, .
     if (!debug_init) {
         debug_init = 1;
         logfile = stderr;
+        debug_mask = DBG_CRIT;
 
         char *env = NULL;
         if ((env = getenv("AACS_DEBUG_MASK")))
